/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>

#ifndef DRVHY8424IPASYN_H
#define DRVHY8424IPASYN_H
#endif

/**************************************************************/
/* A couple of key parameters defining the ADC memory layout. */
#define BANKSIZE  				0x40000        	/* 256K samples per channel for 2MB memory */
#define MAX_MESSAGES  			100           	/* maximum number of messages */
#define MAX_ADC_CHANNELS    	4	      		/* number of ADC channels on 8424 card */
#define HALF_MEM				0x80000			/* 1M Byte or 512K Word*/
#define MAX_CARDS          		256          /* maxiumum number of cards in a system */

/* Identifiers in PROM. */
#define VITA4       			"VITA 4 "
#define HYTECID_LO  			0x0300
#define HYTECID_HI  			0x0080
#define PROM_MODEL  			0x8424
#define NEW_FIRMWARE  			0x8424

/* define bits in the 8424 CSR */
#define CSR_HF      0x0001          /* Set when lower conversion memory filled. */
#define CSR_F       0x0002          /* Set when upper conversion memory filled. */     
#define CSR_CC      0x0004          /* Set when requested sample count captured. */
#define CSR_MIS     0x0008          /* Set when memory update is inhibited. */
#define CSR_MII     0x0010          /* Set if interrupt generated by inhibit (MIS). */
#define CSR_EII     0x0020          /* Set to enable MIS interrupt. */
#define CSR_DA      0x0040          /* Set to allow the unit to disarm on completion of memory acquisition */
#define CSR_1M      0x0080          /* 1 => 128K samples, 0 => 256K samples per channel. */
#define CSR_HE      0x0100          /* Enable interrupt on Half Full set. */
#define CSR_FE      0x0200          /* Enable interrupt on Full set. */
#define CSR_EE      0x0400          /* Enable end of sample sequence interrupt. */
#define CSR_ET      0x0800          /* Allow external enable to control ADC. */
#define CSR_XC      0x1000          /* 0 => use internal clock, 1 => external clock. */
#define CSR_ST      0x2000          /* Set to trigger capture of programmed samples. */
#define CSR_EX      0x4000          /* 0 => continuous capture, 1 => on trigger. */
#define CSR_ARM     0x8000          /* Set to enable capture into waveform RAM. */

#define EXTCSR_TEN  0x0001          /* Triggered status */
#define EXTCSR_SMI  0x0002          /* Software inhibit control and status. Write 1 to inhibit memory writing; write 0 to allow. Read 1 means hardware inhibit; read 0 means hardware allows */
#define EXTCSR_IS   0x0004          /* Single conversion completed flag */
#define EXTCSR_ISE  0x0008          /* Single conversion completed interrupt enable */
#define EXTCSR_CONT 0x0010          /* Allow multiple triggers on CC mode */
#define EXTCSR_R  	0x0040          /* ADC range 0 = +/-10V and 1 = +/-5V */
#define EXTCSR_TE  	0x1000          /* Set 0 = external trigger on rising edge, 1 = external trigger on falling edge */
#define EXTCSR_BTE 	0x2000          /* Set 1 = extrenal trigger on both edge */
#define EXTCSR_CAL	0x4000          /* Set 1 = do not use firmware calibration, 0 = use firmware calibration. No meaning when EX = 1 */

#define TRCSR_TR  	0x8000          /* Transient recorder enable */
#define TRCSR_VT  	0x4000          /* Transient recorder voltage trigger mode */
#define TRCSR_SOL  	0x0020          /* Transient recorder trigger output edge 0=high, 1=low */
#define TRCSR_SIL  	0x0010          /* Transient recorder stop in edge. 0=high, 1=low */
#define TRCSR_ST  	0x0008          /* Transient recorder stop status */
#define TRCSR_SS  	0x0004          /* Transient recorder software stop bit */
#define TRCSR_CHL  	0x0003          /* Choice a channel for VT mode */

#define TRDELAYDIVIDER  		10      /* delay divider for transient recorder */

#define ENABLEADC 				1
#define DISABLEADC 				0

#define INIT 					0
#define RUNTIME 				1

/* driver support flags */
/* Bit0 - continuos mode support */
/* Bit1 - trigger mode support */
#define DRIVER_SUPPORT     		0x0003   		/* support continuous and trigger modes */

#define WAITING_TRIGGER         0           	/* waiting for trigger */
#define TRIGGERRED_COLLECTING   1           	/* triggerred and collecting data */
#define FINISHED_AVAILABLE      2           	/* finished and data available */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif

#define SAMPLEWORDS				2

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* ADC Register and P#define hy8424SupportString			"SUPPORT"
ROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

/* This data structure maps the I/O registers */
typedef struct
{
    USHORT thresholdLow;            /* 0x2C */
    USHORT thresholdHigh;           /* 0x2E */
    USHORT ptrLow[2];        		/* 0x30 - 0x32 */
    USHORT ptrHigh[2];       		/* 0x34 - 0x36 */
    USHORT peak;					/* 0x38 */

}TRANSIENT;

typedef struct
{
    USHORT status;                  /* 0x00 */
    USHORT pointer[2];              /* 0x02 - 0x04 */
    USHORT clockRate;               /* 0x06 */
    USHORT vector;                  /* 0x08 */
    USHORT extStatus;               /* 0x0A */
    USHORT numConversions[2];       /* 0x0C - 0x0E */
    USHORT data[MAX_ADC_CHANNELS];  /* 0x10 - 0x16 */
    USHORT unused1[3]; 				/* 0x18 - 0x1C */
    USHORT average;  				/* 0x1E */
    USHORT unused2[5]; 				/* 0x20 - 0x28 */
    USHORT trstatus;                /* 0x2A */
	TRANSIENT tr[MAX_ADC_CHANNELS];	/* 0x2C - 0x62 */
    USHORT trdelay;             	/* 0x64 */
    USHORT buffer;             		/* 0x66 */
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;            /* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */

    USHORT calType;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
    USHORT unused2;                 /* 0x9C */
    USHORT WLO;                     /* 0x9E */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x2F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x30 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
} IO_MEMORY;

/* enumerated type of the driver's operational modes */
typedef enum{		
	hy8424Continuous,
	hy8424Triggered,
	hy8424Gated,
	hy8424TRSimple,
	hy8424TRVT
} hy8424Modes;


/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

extern "C" {
static void rebootCallbackC     ( void *drvPvt );
static void INTSRC              ( int cardNumber ); 	/* Interrupt Service Routine */
static void intQueuedTaskC      ( void *drvPvt );
static void intSweepQueuedTaskC	( void *drvPvt );
}

/** This is the class definition for the Hy8424Asyn class
  */
class Hy8424Asyn : public asynPortDriver {
public:
	Hy8424Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int clockRate, int extClock, int range);

    /* These are the methods that we override from asynPortDriver */
    virtual asynStatus writeInt32(asynUser *pasynUser, epicsInt32 value);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);
    virtual asynStatus readInt32Array(asynUser *pasynUser, epicsInt32 *value,
                                        size_t nElements, size_t *nIn);
    virtual asynStatus getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high);
    virtual asynStatus writeFloat64(asynUser *pasynUser, epicsFloat64 value);
    virtual asynStatus readFloat64(asynUser *pasynUser, epicsFloat64 *value);
    virtual asynStatus readFloat64Array(asynUser *pasynUser, epicsFloat64 *value,
                                        size_t nElements, size_t *nIn);
    virtual void report(FILE *fp, int details);
    
    /* these are specific to Hy8424Asyn class */
    int Hy8424ContMode(int average, int fastADC);	
    int Hy8424TrigMode(int sample, int average, double offset, int fastADC, int triggerEdge);	
    int Hy8424GateMode(int average, double offset, int fastADC);
    
    //int Hy8424TRInit(int channel, double thresholdlow, double thresholdhigh, int delay);	
    int Hy8424SimpleTR(int sample, double delay, double thresholdlow0, double thresholdhigh0, double thresholdlow1, double thresholdhigh1,
						double thresholdlow2, double thresholdhigh2, double thresholdlow3, double thresholdhigh3, int outputEdge, int stopInEdge);	
	int Hy8424VoltageTR(int sample, int VTChannel, double thresholdlow, double thresholdhigh, int buffersize, int outputEdge, int stopInEdge);

    void rebootCallback();
	void intSR(); 				/* Interrupt Service Routine */
	void intQueuedTask();
	void intSweepQueuedTask();

protected:
    int hy8424SweepData_;				//DATA
    #define FIRST_8424_PARAM hy8424SweepData_
    int hy8424FloatSweepData_;			//FLOATDATA
	int hy8424Data_;					//GENDATA
	int hy8424FloatData_;				//FLOATGENDATA
	int hy8424ScanPeriod_;
	int hy8424SetAverage_;
	int hy8424SetSample_;
	int hy8424SetMode_;
	int hy8424SetOffset_;
	int hy8424Enable_;
	int hy8424SoftTrigger_;
	int hy8424AutoReenable_;
	int hy8424ClearBuffer_;
	int hy8424SetClockRate_;
	int hy8424SetExtClock_;

	int hy8424Range_;
	int hy8424FastADC_;
	int hy8424TriggerEdge_;

	int hy8424TRChannel_;
	int hy8424TRBuffer_;
	int hy8424TRThresholdLow_;
	int hy8424TRThresholdHigh_;
	int hy8424TRPeak_;
	int hy8424TROutputEdge_;
	int hy8424TRInputEdge_;
	int hy8424SoftStop_;

	int hy8424AverageOverflow_;
	int hy8424Overflow_;
	int hy8424BufferCount_;
	int hy8424MessageSent_;
	int hy8424MessageFailed_;
	int hy8424TriggerStatus_;
	int hy8424FWVersion_;
	int hy8424DriverVersion_;
	int hy8424Support_;
    #define LAST_8424_PARAM hy8424Support_

private:
	void setupADC               ( int runtime );
	int setClockRate            ( int runtime );
	int setExtClock             ( int runtime );
	int softwareTrigger         ();
	int setSamples				( int runtime );
	int setAverage 				( int runtime );
	int setOffset				( int runtime );
	int setMode           		( int runtime );
	int setRange           		( int runtime );
	int setEnableDisable		( int enable );

	int setTRLowThreshold       ( int runtime, int channel );
	int setTRHighThreshold      ( int runtime, int channel );
	int setTRChannel			( int runtime );
	int setTRBuffer				( int runtime );
	int setTROutEdge			( int runtime );
	int setTRInEdge				( int runtime );
	int softwareStop			();
	double calcThresholdRaw		( int value );
	
	double toVoltage            ( double rawval );
	int calcThreshold           ( double value );
    
	volatile IO_MEMORY 	*adc;			   	/* pointer to IP card A16 space registers, ID and calibration data */ 
	volatile USHORT 	*convMemoryBase;   	/* pointer to IP card memory space */

	ULONG 	rawData[MAX_ADC_CHANNELS][BANKSIZE];/* ADC data for I/O Intr Int32Array */
    double 	adcData[MAX_ADC_CHANNELS][BANKSIZE];/* ADC data for I/O Intr Float64Array */
    ULONG 	chanData[MAX_ADC_CHANNELS];    	/* storage for the IP card last channel data values */

	int		scanningRate[32];				/* store the scanning rate number */

	/* Storage for the current Configuration parameters */

    USHORT	carrierNum;		                /* carrier index number when adding carrier */
    USHORT	ipSlotNum;		                /* IP card slot number on IP carrier - boot time configuration only */
    USHORT	vectorNum;		                /* Interrupt vector number - boot time configuration only */

    int 	mode;							/* internal mode for ISR */
    int 	enable;							/* internal enable for ISR */
    int 	samples;						/* internal samples for ISR */
    int 	fastadc;						/* internal fastadc for ISR */

    int 	firstTimeFlag;					/* for synchronise the output records */

	int		banksize;
	int 	lastPointer;
    int     sampleCounter;                  /* counter for continuous mode to count samples acquired in order to generate callbacks */

	int 	lastLowPointer;					/* for transient recorder periodical records low */
	int 	lastHighPointer;				/* for transient recorder periodical records high */

/* driver specific parameters */
    int		rebooting;		                /* Flag to indicate is the IOC is currently re-booting */

/* Work Queue Stuff */
    epicsMessageQueueId intMsgQId;
    int messagesSent;                       /* for report */
    int messagesFailed;                     /* for report */
    epicsMessageQueueId intSweepQId;
};

#define hy8424SweepDataString	    "DATA"				//for continuous, trigger and gated mode, this returns sweep data. 
#define hy8424FloatSweepDataString	"FLOATDATA"			//    for transient recorder modes, this returns the data interested
#define hy8424DataString	        "GENDATA"
#define hy8424FloatDataString	    "FLOATGENDATA"
#define hy8424ScanPeriodString	    "SCAN_PERIOD"
#define hy8424SetAverageString	    "SETAVERAGE"
#define hy8424SetSampleString	    "SETSAMPLE"
#define hy8424SetModeString		    "SETMODE"
#define hy8424SetOffsetString	    "SETOFFSET"
#define hy8424EnableString		    "ENABLE"
#define hy8424SoftTriggerString		"SOFTTRIGGER"
#define hy8424AutoReenableString	"REENABLE"
#define hy8424ClearBufferString		"CLEARBUFFER"
#define hy8424SetClockRateString	"SETCLOCKRATE"
#define hy8424SetExtClockString		"SETEXTCLOCK"

#define hy8424RangeString			"RANGE"
#define hy8424FastADCString			"FASTADC"
#define hy8424TriggerEdgeString		"TRIGEDGE"

#define hy8424TRChannelString		"TRCHANNEL"
#define hy8424TRBufferString		"BUFFERSIZE"
#define hy8424TRThresholdLowString	"TRTHRESHOLDLOW"
#define hy8424TRThresholdHighString	"TRTHRESHOLDHIGH"
#define hy8424TRPeakString			"TRPEAK"
#define hy8424TROutputEdgeString	"TROUTEDGE"
#define hy8424TRInputEdgeString		"TRINEDGE"
#define hy8424SoftStopString		"SOFTSTOP"

#define hy8424AverageOverflowString	"AVERAGEOVERFLOW"
#define hy8424OverflowString		"OVERFLOW"
#define hy8424BufferCountString		"BUFFERCOUNT"
#define hy8424MessageSentString		"MSGSENT"
#define hy8424MessageFailedString	"MSGFAILED"
#define hy8424TriggerStatusString	"TRIGGERSTATE"
#define hy8424FWVersionString		"FWVERSION"
#define hy8424DriverVersionString	"DRIVERVERSION"
#define hy8424SupportString			"SUPPORT"

#define NUM_PARAMS (&LAST_8424_PARAM - &FIRST_8424_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (has to be none zero),
 *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
 *				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
 *				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
 *				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz,
 *				  16 = 200khz, 17 = 500Khz, 18 = 1MHz.
 *     extClock = 0 = internal, 1 = extrenal,
 *		  range = 0 : +/-10V, 1 : +/-5V

 *         mode = 0 = continuous, 1 = triggered, 2 = gated, 3 = Transient recorder simple, 4 = transient recorder voltage trigger
 */
extern "C" int Hy8424AsynInit(char *portName, int vmeSlotNum, int ipSlotNum, int vectorNum, 
		int clockRate, int extClock, int range)
{
    Hy8424Asyn *pHy8424Asyn = new Hy8424Asyn(portName, vmeSlotNum, ipSlotNum, vectorNum, clockRate, extClock, range);
    pHy8424Asyn = NULL;  /* This is just to avoid compiler warnings */
    return(asynSuccess);
}

/** Continuous Mode, called directly or from iocsh */
/* Where: -
 *     portName = asyn port name
 *		average = number of readings to SUM for average
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 */
extern "C" int Hy8424ContinueMode(char *portName, int average, int fastADC)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424ContMode(average, fastADC);
    return(asynSuccess);
} 

/** Trigger Mode, called directly or from iocsh 
 *
 * Where: -
 *     portName = asyn port name
 *      samples = number of samples post trigger 
 *		average = number of readings to SUM for average
 *       offset = Offset to start collecting after trigger.
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 *	triggerEdge = 0 : rising edge; 1 : falling edge; 2 : both edge
 */
extern "C" int Hy8424TriggerMode(char *portName, int samples, int average, int offset, 
		int fastADC, int triggerEdge)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424TrigMode(samples, average, (double) offset, fastADC, triggerEdge);
    return(asynSuccess);
} 

/** Gated Mode, called directly or from iocsh 
 *
 * for transient recorder modes, this returns the data interested
 * Where: -
 *     portName = asyn port name
 *		average = number of readings to SUM for average
 *       offset = Offset from the start (or end if negative) of the gated period to the ADC readings to return
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 */
extern "C" int Hy8424GatedMode(char *portName, int average, int offset, int fastADC)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynExtInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynExtInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424GateMode(average, (double) offset, fastADC);
    return(asynSuccess);
} 


/** Simple transient recorder mode, called directly or from iocsh */
/* 
 * Where: -
 *      portName = asyn port name
 *		  sample = number of samples post trigger
 *         delay = delay time from the trigger to the start point of acquisition
 * thresholdlow0 = First threshold point for channel 0. In voltage unit
 *thresholdhigh0 = Second threshold point for channel 0. In voltage unit
 * thresholdlow1 = First threshold point for channel 1. In voltage unit
 *thresholdhigh1 = Second threshold point for channel 1. In voltage unit
 * thresholdlow2 = First threshold point for channel 2. In voltage unit
 *thresholdhigh2 = Second threshold point for channel 2. In voltage unit
 * thresholdlow3 = First threshold point for channel 3. In voltage unit
 *thresholdhigh3 = Second threshold point for channel 3. In voltage unit
 *    outputEdge = Start Out TTL level. 0=high, 1=low
 *    stopInEdge = Stop in TTL level: 0=high, 1=low
 */
extern "C" int Hy8424SimpleTRMode(char *portName, int samples, double delay, double thresholdlow0, 
		double thresholdhigh0, double thresholdlow1, double thresholdhigh1,	double thresholdlow2, 
		double thresholdhigh2, double thresholdlow3, double thresholdhigh3, int outputEdge, int stopInEdge)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynTRInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynTRInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424SimpleTR(samples, delay, thresholdlow0, thresholdhigh0, thresholdlow1, thresholdhigh1, 
    		thresholdlow2, thresholdhigh2, thresholdlow3, thresholdhigh3, outputEdge, stopInEdge);
    return(asynSuccess);
} 

/** Simple transient recorder mode, called directly or from iocsh */
/* 
 * Where: -
 *      portName = asyn port name
 *		  sample = number of samples post trigger
 *         delay = delay time from the trigger to the start point of acquisition
 * thresholdlow0 = First threshold point for channel 0. In voltage unit
 *thresholdhigh0 = Second threshold point for channel 0. In voltage unit
 * thresholdlow1 = First threshold point for channel 1. In voltage unit
 *thresholdhigh1 = Second threshold point for channel 1. In voltage unit
 * thresholdlow2 = First threshold point for channel 2. In voltage unit
 *thresholdhigh2 = Second threshold point for channel 2. In voltage unit
 * thresholdlow3 = First threshold point for channel 3. In voltage unit
 *thresholdhigh3 = Second threshold point for channel 3. In voltage unit
 *    outputEdge = Start Out TTL level. 0=high, 1=low
 *    stopInEdge = Stop in TTL level: 0=high, 1=low
 */
extern "C" int Hy8424SimpleTRMode_vxWorks(char *portName, int samples, int delay, int thresholdlow0, 
		int thresholdhigh0, int thresholdlow1, int thresholdhigh1,	int thresholdlow2, 
		int thresholdhigh2, int thresholdlow3, int thresholdhigh3, int outputEdge, int stopInEdge)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynTRInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynTRInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424SimpleTR(samples, (double)delay, (double)thresholdlow0, (double)thresholdhigh0, (double)thresholdlow1, (double)thresholdhigh1, 
    		(double)thresholdlow2, (double)thresholdhigh2, (double)thresholdlow3, (double)thresholdhigh3, outputEdge, stopInEdge);
    return(asynSuccess);
} 

/** Voltage trigger transient recorder mode, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *		 sample = number of samples post trigger
 *    VTChannel = voltage trigger channel. 0~3
 * thresholdlow = First threshold point for VTChannel . In voltage unit
 *thresholdhigh = Second threshold point for VTChannel. In voltage unit
 *       buffer = Circular buffer size. 0~65535
 *   outputEdge = Start Out TTL level. 0=high, 1=low
 *   stopInEdge = Stop in TTL level: 0=high, 1=low
 */
extern "C" int Hy8424VoltageTRMode(char *portName, int sample, int VTChannel, double thresholdlow, double thresholdhigh, int buffer, int outputEdge, int stopInEdge)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynTRInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynTRInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424VoltageTR(sample, VTChannel, thresholdlow, thresholdhigh, buffer, outputEdge, stopInEdge);
    return(asynSuccess);
} 

/** Voltage trigger transient recorder mode, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *		 sample = number of samples post trigger
 *    VTChannel = voltage trigger channel. 0~3
 * thresholdlow = First threshold point for VTChannel . In voltage unit
 *thresholdhigh = Second threshold point for VTChannel. In voltage unit
 *       buffer = Circular buffer size. 0~65535
 *   outputEdge = Start Out TTL level. 0=high, 1=low
 *   stopInEdge = Stop in TTL level: 0=high, 1=low
 */
extern "C" int Hy8424VoltageTRMode_vxWorks(char *portName, int sample, int VTChannel, int thresholdlow, int thresholdhigh, int buffer, int outputEdge, int stopInEdge)
{
    asynStatus status;
    asynInterface *pasynInterface;
    asynUser *pasynUser;
    Hy8424Asyn *pHy8424 = NULL;

    pasynUser = pasynManager->createAsynUser(0, 0);
    status = pasynManager->connectDevice(pasynUser, portName, 0);
    if (status != asynSuccess) {
        errlogPrintf("Hy8424AsynTRInit, error in connectDevice %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pasynInterface = pasynManager->findInterface(pasynUser, asynFloat64Type, 1);
    if (!pasynInterface) {
        errlogPrintf("Hy8424AsynTRInit cannot find Float64 interface %s\n",
                     pasynUser->errorMessage);
        return asynError;
    }
    pHy8424 = (Hy8424Asyn*) pasynInterface->drvPvt;
    pHy8424->Hy8424VoltageTR(sample, VTChannel, (double)thresholdlow, (double)thresholdhigh, buffer, outputEdge, stopInEdge);
    return(asynSuccess);
} 

// Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "ipSlotNum",iocshArgInt};
static const iocshArg initArg3 = { "vectorNum",iocshArgInt};
static const iocshArg initArg4 = { "clockRate",iocshArgInt};
static const iocshArg initArg5 = { "External Clock", iocshArgInt};
static const iocshArg initArg6 = { "Signal Range", iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4,
                                             &initArg5,
                                             &initArg6};
static const iocshFuncDef initFuncDef = {"Hy8424AsynInit",7,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    Hy8424AsynInit(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival, args[6].ival);
}

// Continue mode init
static const iocshArg initArg0Cont = { "Port name",iocshArgString};
static const iocshArg initArg1Cont = { "average",iocshArgInt};
static const iocshArg initArg2Cont = { "fastADC",iocshArgInt};
static const iocshArg * const initArgsCont[] = {&initArg0Cont,
                                             &initArg1Cont,
                                             &initArg2Cont};
static const iocshFuncDef initFuncDefCont = {"Hy8424ContinueMode",3,initArgsCont};
static void initCallFuncCont(const iocshArgBuf *args)
{
	Hy8424ContinueMode(args[0].sval, args[1].ival, args[2].ival);
}

// Trigger mode init
static const iocshArg initArg0Trig = { "Port name",iocshArgString};
static const iocshArg initArg1Trig = { "samples",iocshArgInt};
static const iocshArg initArg2Trig = { "average",iocshArgInt};
static const iocshArg initArg3Trig = { "offset",iocshArgInt};
static const iocshArg initArg4Trig = { "fastADC", iocshArgInt};
static const iocshArg initArg5Trig = { "Ext Trigger Edge",iocshArgInt};
static const iocshArg * const initArgsTrig[] = {&initArg0Trig,
                                             &initArg1Trig,
                                             &initArg2Trig,
                                             &initArg3Trig,
                                             &initArg4Trig,
                                             &initArg5Trig};
static const iocshFuncDef initFuncDefTrig = {"Hy8424TriggerMode",6,initArgsTrig};
static void initCallFuncTrig(const iocshArgBuf *args)
{
	Hy8424TriggerMode(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
					args[4].ival, args[5].ival);
}


// Gated mode init
static const iocshArg initArg0Gate = { "Port name",iocshArgString};
static const iocshArg initArg1Gate = { "average",iocshArgInt};
static const iocshArg initArg2Gate = { "offset",iocshArgInt};
static const iocshArg initArg3Gate = { "fastADC", iocshArgInt};
static const iocshArg * const initArgsGate[] = {&initArg0Gate,
                                             &initArg1Gate,
                                             &initArg2Gate,
                                             &initArg3Gate};
static const iocshFuncDef initFuncDefGate = {"Hy8424GatedMode",4,initArgsGate};
static void initCallFuncGate(const iocshArgBuf *args)
{
	Hy8424GatedMode(args[0].sval, args[1].ival, args[2].ival, args[3].ival);
}

// Simple transient recorder mode init
static const iocshArg initArg0SIMTR = { "Port name",iocshArgString};
static const iocshArg initArg1SIMTR = { "samples",iocshArgInt};
static const iocshArg initArg2SIMTR = { "delay",iocshArgDouble};
static const iocshArg initArg3SIMTR = { "ThresholdLow0",iocshArgDouble};
static const iocshArg initArg4SIMTR = { "ThresholdHigh0",iocshArgDouble};
static const iocshArg initArg5SIMTR = { "ThresholdLow1",iocshArgDouble};
static const iocshArg initArg6SIMTR = { "ThresholdHigh1", iocshArgDouble};
static const iocshArg initArg7SIMTR = { "ThresholdLow2",iocshArgDouble};
static const iocshArg initArg8SIMTR = { "ThresholdHigh2", iocshArgDouble};
static const iocshArg initArg9SIMTR = { "ThresholdLow3",iocshArgDouble};
static const iocshArg initArg10SIMTR = { "ThresholdHigh3", iocshArgDouble};
static const iocshArg initArg11SIMTR = { "OutputEdge",iocshArgInt};
static const iocshArg initArg12SIMTR = { "StopInEdge",iocshArgInt};
static const iocshArg * const initArgsSIMTR[] = {&initArg0SIMTR,
                                             &initArg1SIMTR,
                                             &initArg2SIMTR,
                                             &initArg3SIMTR,
                                             &initArg4SIMTR,
                                             &initArg5SIMTR,
                                             &initArg6SIMTR,
                                             &initArg7SIMTR,
                                             &initArg8SIMTR,
                                             &initArg9SIMTR,
                                             &initArg10SIMTR,
                                             &initArg11SIMTR,
                                             &initArg12SIMTR};
static const iocshFuncDef initFuncDefSIMTR = {"Hy8424SimpleTRMode",13,initArgsSIMTR};
static void initCallFuncSIMTR(const iocshArgBuf *args)
{
    Hy8424SimpleTRMode(args[0].sval, args[1].ival, args[2].dval, args[3].dval,
			args[4].dval, args[5].dval, args[6].dval, args[7].dval,	args[8].dval,
			args[9].dval, args[10].dval, args[11].ival, args[12].ival);
}

static const iocshArg initArg0SIMTR_vx = { "Port name",iocshArgString};
static const iocshArg initArg1SIMTR_vx = { "samples",iocshArgInt};
static const iocshArg initArg2SIMTR_vx = { "delay",iocshArgInt};
static const iocshArg initArg3SIMTR_vx = { "ThresholdLow0",iocshArgInt};
static const iocshArg initArg4SIMTR_vx = { "ThresholdHigh0",iocshArgInt};
static const iocshArg initArg5SIMTR_vx = { "ThresholdLow1",iocshArgInt};
static const iocshArg initArg6SIMTR_vx = { "ThresholdHigh1", iocshArgInt};
static const iocshArg initArg7SIMTR_vx = { "ThresholdLow2",iocshArgInt};
static const iocshArg initArg8SIMTR_vx = { "ThresholdHigh2", iocshArgInt};
static const iocshArg initArg9SIMTR_vx = { "ThresholdLow3",iocshArgInt};
static const iocshArg initArg10SIMTR_vx = { "ThresholdHigh3", iocshArgInt};
static const iocshArg initArg11SIMTR_vx = { "OutputEdge",iocshArgInt};
static const iocshArg initArg12SIMTR_vx = { "StopInEdge",iocshArgInt};
static const iocshArg * const initArgsSIMTR_vx[] = {&initArg0SIMTR_vx,
                                             &initArg1SIMTR_vx,
                                             &initArg2SIMTR_vx,
                                             &initArg3SIMTR_vx,
                                             &initArg4SIMTR_vx,
                                             &initArg5SIMTR_vx,
                                             &initArg6SIMTR_vx,
                                             &initArg7SIMTR_vx,
                                             &initArg8SIMTR_vx,
                                             &initArg9SIMTR_vx,
                                             &initArg10SIMTR_vx,
                                             &initArg11SIMTR_vx,
                                             &initArg12SIMTR_vx};
static const iocshFuncDef initFuncDefSIMTR_vx = {"Hy8424SimpleTRMode",13,initArgsSIMTR_vx};
static void initCallFuncSIMTR_vx(const iocshArgBuf *args)
{
    Hy8424SimpleTRMode_vxWorks(args[0].sval, args[1].ival, args[2].ival, args[3].ival,
			args[4].ival, args[5].ival, args[6].ival, args[7].ival,	args[8].ival,
			args[9].ival, args[10].ival, args[11].ival, args[12].ival);
}

// Transient recorder extinit
static const iocshArg initArg0VOLTR = { "Port name",iocshArgString};
static const iocshArg initArg1VOLTR = { "samples",iocshArgInt};
static const iocshArg initArg2VOLTR = { "VTChannel",iocshArgInt};
static const iocshArg initArg3VOLTR = { "ThresholdLow",iocshArgDouble};
static const iocshArg initArg4VOLTR = { "ThresholdHigh",iocshArgDouble};
static const iocshArg initArg5VOLTR = { "Buffer",iocshArgInt};
static const iocshArg initArg6VOLTR = { "OutputEdge",iocshArgInt};
static const iocshArg initArg7VOLTR = { "StopInEdge",iocshArgInt};
static const iocshArg * const initArgsVOLTR[] = {&initArg0VOLTR,
                                             	&initArg1VOLTR,
                                             	&initArg2VOLTR,
                                             	&initArg3VOLTR,
                                             	&initArg4VOLTR,
                                             	&initArg5VOLTR,
                                             	&initArg6VOLTR,
                                             	&initArg7VOLTR};
static const iocshFuncDef initFuncDefVOLTR = {"Hy8424VoltageTRMode",8,initArgsVOLTR};
static void initCallFuncVOLTR(const iocshArgBuf *args)
{
    Hy8424VoltageTRMode(args[0].sval, args[1].ival, args[2].ival, args[3].dval, args[4].dval, args[5].ival,  args[6].ival, args[7].ival);
}

static const iocshArg initArg0VOLTR_vx = { "Port name",iocshArgString};
static const iocshArg initArg1VOLTR_vx = { "samples",iocshArgInt};
static const iocshArg initArg2VOLTR_vx = { "VTChannel",iocshArgInt};
static const iocshArg initArg3VOLTR_vx = { "ThresholdLow",iocshArgInt};
static const iocshArg initArg4VOLTR_vx = { "ThresholdHigh",iocshArgInt};
static const iocshArg initArg5VOLTR_vx = { "Buffer",iocshArgInt};
static const iocshArg initArg6VOLTR_vx = { "OutputEdge",iocshArgInt};
static const iocshArg initArg7VOLTR_vx = { "StopInEdge",iocshArgInt};
static const iocshArg * const initArgsVOLTR_vx[] = {&initArg0VOLTR_vx,
                                             	&initArg1VOLTR_vx,
                                             	&initArg2VOLTR_vx,
                                             	&initArg3VOLTR_vx,
                                             	&initArg4VOLTR_vx,
                                             	&initArg5VOLTR_vx,
                                             	&initArg6VOLTR_vx,
                                             	&initArg7VOLTR_vx};
static const iocshFuncDef initFuncDefVOLTR_vx = {"Hy8424VoltageTRMode",8,initArgsVOLTR_vx};
static void initCallFuncVOLTR_vx(const iocshArgBuf *args)
{
    Hy8424VoltageTRMode_vxWorks(args[0].sval, args[1].ival, args[2].ival, args[3].ival, args[4].ival, args[5].ival,  args[6].ival, args[7].ival);
}



void drvHy8424AsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
    iocshRegister(&initFuncDefCont,initCallFuncCont);
    iocshRegister(&initFuncDefTrig,initCallFuncTrig);
    iocshRegister(&initFuncDefGate,initCallFuncGate);
    iocshRegister(&initFuncDefSIMTR,initCallFuncSIMTR);
    iocshRegister(&initFuncDefSIMTR_vx,initCallFuncSIMTR_vx);
    iocshRegister(&initFuncDefVOLTR,initCallFuncVOLTR);
    iocshRegister(&initFuncDefVOLTR_vx,initCallFuncVOLTR_vx);
}

extern "C" {
epicsExportRegistrar(drvHy8424AsynRegister);
}

